import pywapor
from sentinelsat import SentinelAPI
import os
from datetime import datetime as dt
import tqdm
import shutil
import pywapor
import warnings
from pywapor.general.processing_functions import save_ds, open_ds, create_wkt, unpack, make_example_ds, remove_ds, adjust_timelim_dtype
from pywapor.general.logger import log
from pywapor.enhancers.apply_enhancers import apply_enhancer
import xarray as xr
import rioxarray.merge
import tqdm
import numpy as np
import glob
import rasterio.crs
import types
import logging
from sentinelsat.download import Downloader

def download(folder, latlim, lonlim, timelim, search_kwargs, node_filter = None):

    if not os.path.isdir(folder):
        os.makedirs(folder)

    timelim = adjust_timelim_dtype(timelim)

    def _progress_bar(self, **kwargs):
        if "checksumming" in kwargs.get("desc", "no_desc"):
            kwargs.update({"disable": True, "delay": 15})
        elif "Fetching" in kwargs.get("desc", "no_desc"):
            kwargs.update({"disable": True, "delay": 15})
        elif "Downloading products" in kwargs.get("desc", "no_desc"):
            kwargs.update({"disable": True, "position": 0, "desc": "Downloading scenes", "unit": "scene"})
        else:
            kwargs.update({"disable": False, "position": 0})
        kwargs.update({"leave": False})
        return tqdm.tqdm(**kwargs)

    un, pw = pywapor.collect.accounts.get('SENTINEL')
    api = SentinelAPI(un, pw, 'https://apihub.copernicus.eu/apihub')
    api._tqdm = types.MethodType(_progress_bar, api)
    
    logger = api.logger
    logger.setLevel('INFO')
    handler = logging.FileHandler(filename = os.path.join(folder, "log_sentinelapi.txt"))
    formatter = logging.Formatter(fmt='%(levelname)s %(asctime)s: %(message)s')
    handler.setLevel('INFO')
    handler.setFormatter(formatter)
    logger.addHandler(handler)

    footprint = create_wkt(latlim, lonlim)
    products = api.query(footprint, date = tuple(timelim), **search_kwargs)
    log.info(f"--> Found {len(products)} {search_kwargs['producttype']} scenes.")

    dler = Downloader(api, node_filter = node_filter)
    dler._tqdm = types.MethodType(_progress_bar, dler)

    statuses, exceptions, out = dler.download_all(products, folder)

    if len(exceptions) > 0:
        log.info(f"--> An exception occured, check `log_sentinelapi.txt` for more info.")
 
    log.info(f"--> Finished downloading {search_kwargs['producttype']} scenes.")

    if isinstance(node_filter, type(None)):
        scenes = [x["path"] for x in out.values()]
    else:
        scenes = [os.path.join(folder, x["node_path"][2:]) for x in out.values()]
    
    return scenes

def process_sentinel(scenes, variables, source_name, time_func, final_fn, post_processors, bb = None):

    chunks = {"time": 1, "x": 1000, "y": 1000}

    example_ds = None
    dss1 = dict()

    log.info(f"--> Processing {len(scenes)} scenes.").add()

    target_crs = rasterio.crs.CRS.from_epsg(4326)

    for i, scene_folder in enumerate(scenes):
        
        folder, fn = os.path.split(scene_folder)

        ext = os.path.splitext(scene_folder)[-1]
        
        fp = os.path.join(folder, os.path.splitext(fn)[0] + ".nc")
        if os.path.isfile(fp):
            ds = open_ds(fp)
            dtime = ds.time.values[0]
            if dtime in dss1.keys():
                dss1[dtime].append(ds)
            else:
                dss1[dtime] = [ds]
            continue

        if ext == ".zip":
            scene_folder = unpack(fn, folder)
            remove_folder = True
        else:
            scene_folder = scene_folder
            remove_folder = False

        if source_name == "SENTINEL2":
            ds = pywapor.collect.product.SENTINEL2.s2_processor(scene_folder, variables)
        elif source_name == "SENTINEL3":
            ds = pywapor.collect.product.SENTINEL3.s3_processor(scene_folder, variables, bb = bb)
        else:
            raise ValueError

        # Apply variable specific functions.
        for vars in variables.values():
            for func in vars[2]:
                ds, label = apply_enhancer(ds, vars[1], func)

        # NOTE: see https://github.com/corteva/rioxarray/issues/545
        # NOTE: see rioxarray issue here: https://github.com/corteva/rioxarray/issues/570
        ds = ds.sortby("y", ascending = False)
        _ = [ds[var].rio.write_nodata(np.nan, inplace = True) for var in ds.data_vars]

        # Clip and pad to bounding-box
        if isinstance(example_ds, type(None)):
            example_ds = make_example_ds(ds, folder, target_crs, bb = bb)
        ds = ds.rio.reproject_match(example_ds).chunk({k: v for k, v in chunks.items() if k in ["x", "y"]})

        # NOTE: see rioxarray issue here: https://github.com/corteva/rioxarray/issues/570
        _ = [ds[var].attrs.pop("_FillValue") for var in ds.data_vars if "_FillValue" in ds[var].attrs.keys()]
        
        ds = ds.assign_coords({"x": example_ds.x, "y": example_ds.y})

        dtime = time_func(fn)
        ds = ds.expand_dims({"time": 1})
        ds = ds.assign_coords({"time":[dtime]})

        # Save to netcdf
        ds = save_ds(ds, fp, chunks = chunks, encoding = "initiate", label = f"({i+1}/{len(scenes)}) Processing {fn} to netCDF.")

        if dtime in dss1.keys():
            dss1[dtime].append(ds)
        else:
            dss1[dtime] = [ds]

        if remove_folder:
            shutil.rmtree(scene_folder)

    log.sub()
    
    # Merge spatially.
    dss = [xr.concat(dss0, "stacked").median("stacked") for dss0 in dss1.values()]

    # Merge temporally.
    ds = xr.concat(dss, "time")

    # Define output path.
    fp = os.path.join(folder, final_fn)
    if os.path.isfile(fp):
        os.remove(fp)
    
    # Apply general product functions.
    for var, funcs in post_processors.items():
        for func in funcs:
            ds, label = apply_enhancer(ds, var, func)
            log.info(label)

    # Remove unrequested variables.
    ds = ds[list(post_processors.keys())]
    
    for var in ds.data_vars:
        ds[var].attrs = {}

    ds = ds.rio.write_crs(target_crs)

    ds = ds.sortby("time")

    # Save final netcdf.
    with warnings.catch_warnings():
        warnings.filterwarnings("ignore", message="invalid value encountered in true_divide")
        warnings.filterwarnings("ignore", message="divide by zero encountered in true_divide")
        ds = save_ds(ds, fp, chunks = chunks, encoding = "initiate", label = f"Merging files.")

    # Remove intermediate files.
    for dss0 in dss1.values():
        for x in dss0:
            remove_ds(x)

    return ds

# if __name__ == "__main__":

    # folder = r"/Users/hmcoerver/On My Mac/sentinel_dl_test/SENTINEL3"
    # latlim = [28.9, 29.7]
    # lonlim = [30.2, 31.2]
    # timelim = ["2022-06-01", "2022-07-11"]
    # product_name = 'SL_2_LST___'
    # node_filter = None

    # search_kwargs = {
    #                     "platformname": "Sentinel-3",
    #                     "producttype": product_name,
    #                     "limit": 5,
    #                     }

    # scenes = download(folder, latlim, lonlim, timelim, search_kwargs, node_filter = node_filter)

    # import numpy as np

    # folder = r"/Users/hmcoerver/On My Mac/sentinel_dl_test/SENTINEL2"
    # latlim = [28.9, 29.7]
    # lonlim = [30.2, 31.2]
    # timelim = ["2022-06-01", "2022-07-11"]
    # product_name = 'S2MSI2A'
    # req_vars = ["ndvi", "r0"]

    # adjust_logger(True, folder, "INFO")

    # variables = pywapor.collect.product.SENTINEL2.default_vars(product_name, req_vars)

    # def node_filter(node_info):
    #     fn = os.path.split(node_info["node_path"])[-1]
    #     to_dl = list(variables.keys())
    #     return np.any([x in fn for x in to_dl])

    # search_kwargs = {
    #                 "platformname": "Sentinel-2",
    #                 "producttype": product_name,
    #                 "limit": 5,
    #                 }

    # scenes = download(folder, latlim, lonlim, timelim, search_kwargs, node_filter = node_filter)

    ##

    # if isinstance(timelim[0], str):
    #     timelim[0] = dt.strptime(timelim[0], "%Y-%m-%d")
    #     timelim[1] = dt.strptime(timelim[1], "%Y-%m-%d")

    # def _progress_bar(self, **kwargs):
    #     if "checksumming" in kwargs.get("desc", "no_desc"):
    #         kwargs.update({"disable": True, "delay": 15})
    #     elif "Fetching" in kwargs.get("desc", "no_desc"):
    #         kwargs.update({"disable": True, "delay": 15})
    #     elif "Downloading products" in kwargs.get("desc", "no_desc"):
    #         kwargs.update({"disable": True, "position": 0, "desc": "Downloading scenes", "unit": "scene"})
    #     else:
    #         kwargs.update({"disable": False, "position": 0})
    #     kwargs.update({"leave": False})
    #     return tqdm.tqdm(**kwargs)

    # sentinel_id = 'a2b9a381-3e9a-4fc8-a5ec-4040d43b83b7'
    # search_kwargs.update({"uuid": sentinel_id})
    
    # un, pw = pywapor.collect.accounts.get('SENTINEL')
    # api = SentinelAPI(un, pw, 'https://apihub.copernicus.eu/apihub')
    # api._tqdm = types.MethodType(_progress_bar, api)

    # footprint = create_wkt(latlim, lonlim)
    # products = api.query(footprint, **search_kwargs)
    # log.info(f"--> Found {len(products)} {search_kwargs['producttype']} scenes.")

    # dler = Downloader(api, node_filter = node_filter)
    # dler._tqdm = types.MethodType(_progress_bar, dler)

    # statuses, exceptions, out = dler.download_all(products, folder)



"""
Functions to prepare input for `pywapor.se_root`, more specifically to
interpolate various parameters in time to match with land-surface-temperature
times. 
"""

from pywapor.general.processing_functions import save_ds, open_ds, remove_ds, log_example_ds
from pywapor.general.reproject import align_pixels
from pywapor.enhancers.apply_enhancers import apply_enhancer
from pywapor.general.logger import log
import os
import numpy as np
import xarray as xr
from itertools import chain
import pywapor.general.levels as levels


def is_aligned(ds, example_ds):
    return ds.equals(example_ds)


def main(dss, sources, folder, general_enhancers, example_t_vars = ["lst"]):
    """Aligns the datetimes in de `dss` xr.Datasets with the datetimes of the 
    dataset with variable `example_t_var`.

    Parameters
    ----------
    dss : dict
        Keys are tuples of (`source`, `product_name`), values are xr.Dataset's 
        or paths (str) to netcdf files, which will be aligned along the time dimensions.
    sources : dict
        Configuration for each variable and source.
    folder : str
        Path to folder in which to store (intermediate) data.
    enhancers : list | "default", optional
        Functions to apply to the xr.Dataset before creating the final
        output, by default "default".
    example_t_var : str, optional
        Which variable to align the other datasets to in the time dimension, by default "lst".

    Returns
    -------
    xr.Dataset
        Dataset in which all variables have been interpolated to the same times.
    """
    chunks = {"time": -1, "y": 500, "x": 500}

    # Open unopened netcdf files.
    dss = {**{k: open_ds(v) for k, v in dss.items() if isinstance(v, str)}, 
            **{k:v for k,v in dss.items() if not isinstance(v, str)}}

    # Determine final output path.
    final_path = os.path.join(folder, "se_root_in.nc")

    # Make list to store intermediate datasets.
    dss2 = dict()

    # Make inventory of all variables.
    variables = [x for x in np.unique(list(chain.from_iterable([ds.data_vars for ds in dss.values()]))).tolist() if x in sources.keys()]
    variables = example_t_vars + [x for x in variables if x not in example_t_vars]

    # Create variable to store times to interpolate to.
    example_time = None

    cleanup = list()

    # Loop over the variables
    for var in variables:

        config = sources[var]
        spatial_interp = config["spatial_interp"]
        temporal_interp = config["temporal_interp"]

        # Align pixels of different products for a single variable together.
        dss_part = [ds[[var]] for ds in dss.values() if var in ds.data_vars]
        dss1, temp_files1 = align_pixels(dss_part, folder, spatial_interp, fn_append = "_step1")
        cleanup.append(temp_files1)

        # Combine different source_products (along time dimension).
        ds = xr.combine_nested(dss1, concat_dim = "time").chunk({"time": -1}).sortby("time").squeeze()

        if var in example_t_vars:
            if "time" not in ds[var].dims:
                ds[var] = ds[var].expand_dims("time").transpose("time", "y", "x")
            if isinstance(example_time, type(None)):
                example_time = ds["time"]
            else:
                example_time = xr.concat([example_time, ds["time"]], dim = "time").drop_duplicates("time").sortby("time")
            dss2[var] = ds
        elif "time" in ds[var].dims:
            lbl = f"Aligning times in `{var}` ({ds.time.size}) with `{'` and `'.join(example_t_vars)}` ({example_time.time.size}, {temporal_interp})."
            ds = ds.interpolate_na(dim = "time", method = temporal_interp).ffill("time").bfill("time")
            ds = ds.interp_like(example_time, method = temporal_interp)
            dst_path = os.path.join(folder, f"{var}_i.nc")
            ds = save_ds(ds, dst_path, chunks = chunks, encoding = "initiate", label = lbl)
            dss2[var] = ds
            cleanup.append([ds])
        else:
            # Add time-invariant data as is.
            dss2[var] = ds

    # Apply variable specific enhancers
    variables = levels.find_setting(sources, "variable_enhancers")
    for var in variables:
        for func in sources[var]["variable_enhancers"]:
            dss2 = func(dss2, var, folder)

    # Align all the variables together.
    example_source = levels.find_setting(sources, "is_example", max_length = 1, min_length = 1)
    if len(example_source) == 1:
        example_ds = dss[example_source[0]]
        log_example_ds(example_ds)
    else:
        log.warning(f"--> No valid example dataset set.")
        example_ds = None
    spatial_interps = [sources[list(x.data_vars)[0]]["spatial_interp"] for x in dss2.values()]
    dss3, temp_files3 = align_pixels(dss2.values(), folder, spatial_interps, example_ds, fn_append = "_step2")
    cleanup.append(temp_files3)

    # Merge everything.
    ds = xr.merge(dss3)

    # Apply general enhancers.
    for func in general_enhancers:
        ds, label = apply_enhancer(ds, None, func)
        log.info(label)

    while os.path.isfile(final_path):
        final_path = final_path.replace(".nc", "_.nc")

    if ds.time.size == 0:
        log.warning("--> No valid data created (ds.time.size == 0).")
        return ds

    ds = save_ds(ds, final_path, encoding = "initiate",
                    label = f"Creating merged file `{os.path.split(final_path)[-1]}`.")

    for nc in list(chain.from_iterable(cleanup)):
        remove_ds(nc)

    return ds

# if __name__ == "__main__":

#     import numpy as np
#     import glob

#     dss = {
#         ("SENTINEL2", "S2MSI2A"): r"/Users/hmcoerver/Local/test_data/SENTINEL2/S2MSI2A.nc",
#         ("SENTINEL3", "SL_2_LST___"): r"/Users/hmcoerver/Local/test_data/SENTINEL3/SL_2_LST___.nc",
#         # ("VIIRSL1", "VNP02IMG"): r"/Users/hmcoerver/Local/test_data/VIIRSL1/VNP02IMG.nc",
#     }
#     example_source = ("SENTINEL2", "S2MSI2A")

#     sources = {
#             "ndvi": {"spatial_interp": "nearest", "temporal_interp": "linear"},
#             "lst": {"spatial_interp": "nearest", "temporal_interp": "linear"},
#             "bt": {"spatial_interp": "nearest", "temporal_interp": "linear"},
#                 }

#     # example_source = ("source1", "productX")
#     folder = r"/Users/hmcoerver/Local/test_data"
#     enhancers = list()
#     example_t_vars = ["lst"]

#     chunks = (1, 500, 500)

#     for fp in glob.glob(os.path.join(folder, "*.nc")):
#         os.remove(fp)

#     from pywapor.general.logger import log, adjust_logger
#     adjust_logger(True, folder, "INFO")

#     ds = main(dss, sources, example_source, folder, enhancers, example_t_vars = ["lst"])


import numpy as np
import xarray as xr

def epsilon_soil_moisture(ef_24):
    r"""
    Computes the epsilon soil moisture

    .. math ::
        I = \frac{\lambda I^*}{86400}

    Parameters
    ----------
    ef_24 : float
        evaporative fraction
        :math:`EF_{ref}`
        [-]

    Returns
    -------
    eps_w : float
        epsilon soil moisture
        :math:`I`
        [-]
    """
    eps_w = 0.5 + 0.5 * ef_24
     
    return eps_w

def epsilon_autotrophic_respiration():   
    r"""
    Computes the epsilon autotrophic respiration

    Returns
    -------
    eps_a : float
        epsilon autotrophic respiration
        :math:`I`
        [-]
    """    
    
    # eps_a = 1-0.23 # 23% of GPP
    eps_a = 1-0.5
    
    return eps_a

def stress_radiation(ra_24):
    r"""
    Computes the stress for plants when there is not sufficient radiation

    .. math ::
        S_{r}=\frac{S^{\downarrow}}{\left(S^{\downarrow}+60.\right)}
        \left(1+\frac{60}{500}\right)

    Parameters
    ----------
    ra_24 : float
        daily solar radiation
        :math:`S^{\downarrow}`
        [Wm-2]

    Returns
    -------
    stress_rad : float
        stress factor for radiation
        :math:`S_{r}`
        [-]


    Examples
    --------
    >>> import ETLook.stress as stress
    >>> stress.stress_radiation()
    0.0
    >>> stress.stress_radiation(500)
    1.0
    >>> stress.stress_radiation(700)
    1.0
    >>> stress.stress_radiation(250)
    0.90322580645161288
    """
    stress = ra_24/(ra_24 + 60.)*(1 + 60./500.)
    
    if isinstance(stress, xr.DataArray):
        stress = stress.clip(0, 1)
    else:
        stress = np.clip(stress, 0, 1)

    return stress


def stress_moisture(se_root, tenacity=1.5):
    r"""
    Computes the stress for plants when there is not sufficient soil
    moisture in the root zone

    .. math ::
        S_{m}=K_{sf}S_{e,root}-\frac{\sin\left(2\pi S_{e,root}\right)}{2\pi}

    The tenacity factor :math:`K_{sf}` ranges from 1 for sensitive plants to
    1.5 for moderately sensitive plants to 3 for insensitive
    (tenacious plants).

    Parameters
    ----------
    se_root : float
        effective saturation root zone moisture
        `S_{e,root}`
        [-]
    tenacity : float
        tenacity factor
        `K_{sf}`
        [-]

    Returns
    -------
    stress_moist : float
        stress factor for root zone moisture
        :math:`S_{m}`
        [-]

    Examples
    --------
    >>> import ETLook.stress as stress
    >>> stress.stress_moisture(0.5)
    0.75
    >>> stress.stress_moisture(0.5, tenacity = 1)
    0.5
    >>> stress.stress_moisture(0.5, tenacity = 3)
    1.0
    """
    stress = tenacity*se_root - (np.sin(2*np.pi*se_root))/(2*np.pi)

    if isinstance(stress, xr.DataArray):
        stress = stress.clip(0,1)
    else:
        stress = np.clip(stress, 0, 1)

    return stress


def stress_temperature(t_air_24, t_opt=25.0, t_min=0.0, t_max=50.0):
    r"""
    Computes the stress for plants when it is too cold or hot

    .. math ::
        f=\frac{T_{max}-T_{opt}}{T_{opt}-T_{min}}

    .. math ::
        s_{T}=\frac{\left(T_{a}-T_{min}\right)\left(T_{max}-T_{a}\right)^{f}}
            {\left(T_{opt}-T_{min}\right)\left(T_{max}-T_{opt}\right)^{f}}

    Parameters
    ----------
    t_air_24 : float
        daily air temperature
        :math:`T_{a}`
        [C]
    t_opt : float
        optimum air temperature for plant growth
        :math:`T_{opt}`
        [C]
    t_min : float
        minimum air temperature for plant growth
        :math:`T_{min}`
        [C]
    t_max : float
        maximum air temperature for plant growth
        :math:`T_{max}`
        [C]

    Returns
    -------
    stress_temp : float
        stress factor for air temperature
        :math:`S_{T}`
        [-]

    Examples
    --------
    >>> import ETLook.stress as stress
    >>> stress.stress_temperature(15)
    0.83999999999999997
    >>> stress.stress_temperature(15, t_opt =20)
    0.9451080185178129
    >>> stress.stress_temperature(15, t_opt =20, t_min=10)
    0.79398148148148151
    >>> stress.stress_temperature(15, t_opt =20, t_min=10, t_max=30)
    0.75

    """
    # f = float((t_max - t_opt))/float((t_opt - t_min))
    f = (t_max - t_opt)/(t_opt - t_min)
    x = (t_air_24 - t_min) * (t_max - t_air_24)**f
    y = (t_opt - t_min) * (t_max - t_opt)**f

    stress = x/y

    if isinstance(stress, xr.DataArray):
        stress = stress.clip(0, 1)
    else:
        stress = np.clip(stress, 0, 1)
        
    return stress


def stress_vpd(vpd_24, vpd_slope=-0.3):
    r"""
    Computes the stress for plants if the vpd increases too much. With
    lower slopes the stress increases faster. The slope of the curve
    is between -0.3 and -0.7

    .. math ::
        S_{v}=m\ln(0.1\Delta_{e}+\frac{1}{2})+1

    Parameters
    ----------
    vpd_24 : float
        daily vapour pressure deficit
        :math:`\Delta_{e}`
        [mbar]
    vpd_slope : float
        vapour pressure stress curve slope
        :math:`m`
        [mbar-1]

    Returns
    -------
    stress_vpd : float
        stress factor for vapour pressure deficit
        :math:`S_{v}`
        [-]

    Examples
    --------
    >>> import ETLook.stress as stress
    >>> stress.stress_vpd(15)
    0.79205584583201638
    >>> stress.stress_vpd(15, vpd_slope=-0.7)
    0.51479697360803833
    >>> stress.stress_vpd(15, vpd_slope=-0.3)
    0.79205584583201638

    """
    stress = vpd_slope * np.log(vpd_24/10. + 0.5) + 1

    if isinstance(stress, xr.DataArray):
        stress = stress.clip(0, 1)
    else:
        stress = np.clip(stress, 0, 1)


    return stress



"""Functions that convert different landuse classifications into the ETLook variables
`land_mask`, `lue_max`, `rs_min` and `z_obst_max`.
"""
import numpy as np

def lulc_to_x(ds, var, convertor, in_var = None, out_var = None):
    """Convert values in a classified map to other values based on a dictionary.

    Parameters
    ----------
    ds : xr.Dataset
        Dataset containing `var`.
    var : str
        Variable name.
    convertor : dict
        Dictionary with keys corresponding to values in `var` and values the replacements.
    out_var : str, optional
        New variable name in case `var` should not be overwritten, by default None.

    Returns
    -------
    xr.Dataset
        Dataset with the replaced values.
    """
    if not isinstance(in_var, type(None)):
        var = in_var

    new_data = ds[var] * np.nan

    for class_number in np.unique(ds[var]):
        if class_number in convertor.keys():
            new_data = new_data.where(ds[var] != class_number, 
                                        convertor[class_number])

    if "sources" in ds[var].attrs.keys():
        new_data.attrs = {"sources": ds[var].attrs["sources"]}

    if not isinstance(out_var, type(None)):
        ds[out_var] = new_data
    else: 
        ds[var] = new_data

    return ds

def wapor_to_land_mask():
    """Generate a converter dictionary to be used by `lulc.lulc_to_x` to convert
    from WaPOR-lulc to the ETlook `land_mask`.

    Returns
    -------
    dict
        Describes the conversions to be made.
    """

    convertor = {
                    20: 1,
                    30: 1,
                    42: 1,
                    43: 1,
                    50: 3,
                    60: 1,
                    80: 2,
                    90: 1,
                    116: 1,
                    126 : 1,
                    33 : 1,
                    41 : 1,
                }

    return convertor

def wapor_to_lue_max():
    """Generate a converter dictionary to be used by `lulc.lulc_to_x` to convert
    from WaPOR-lulc to the ETlook `lue_max`.

    Returns
    -------
    dict
        Describes the conversions to be made.
    """

    convertor = {
                    20:	2.10,	# Shrubland
                    30:	2.39,	# Grassland
                    41:	2.7,	# Cropland, rainfed
                    42:	2.7,	# Cropland, irrigated or under water management
                    43:	2.7,	# Cropland, fallow
                    50:	1.42,	# Built-up
                    60:	1.42,	# Bare / sparse vegetation
                    70:	0.0,	# Permament snow / ice
                    80:	0.0,	# Water bodies
                    81:	1.43,	# Temporary water bodies
                    90:	2.1,	# Shrub or herbaceous cover, flooded
                    111: 1.98,	# Tree cover: closed, evergreen needle-leaved
                    112: 2.56,	# Tree cover: closed, evergreen broadleaved
                    114: 1.99,	# Tree cover: closed, deciduous broadleaved
                    115: 2.23,	# Tree cover: closed, mixed type
                    116: 2.23,	# Tree cover: closed, unknown type
                    121: 1.98,	# Tree cover: open, evergreen needle-leaved
                    122: 2.56,	# Tree cover: open, evergreen broadleaved
                    123: 1.57,	# Tree cover: open, deciduous needle-leaved
                    124: 2.48,	# Tree cover: open, deciduous broadleaved
                    125: 2.23,	# Tree cover: open, mixed type
                    126: 2.23,	# Tree cover: open, unknown type
                    200: 0.0,	# Sea water
    }

    return convertor

def wapor_to_rs_min():
    """Generate a converter dictionary to be used by `lulc.lulc_to_x` to convert
    from WaPOR-lulc to the ETlook `rs_min`.

    Returns
    -------
    dict
        Describes the conversions to be made.
    """

    convertor = {
                20 : 175,
                30 : 175,
                42 : 125,
                43 : 125,
                50 : 400,
                60 : 100,
                80 : 100,
                90 : 150,
                116 : 180,
                126 : 250,
                33 : 175,
                41 : 150,
    }

    return convertor

def wapor_to_z_obst_max():
    """Generate a converter dictionary to be used by `lulc.lulc_to_x` to convert
    from WaPOR-lulc to the ETlook `z_obst_max`.

    Returns
    -------
    dict
        Describes the conversions to be made.
    """

    convertor = {
                    20 : 1.0,
                    30 : 2.0,
                    42 : 1.5,
                    43 : 1.5,
                    50 : 10.0,
                    60 : 0.1,
                    80 : 0.1,
                    90 : 2.0,
                    116 : 5.0,
                    126 : 3.0,
                    33 : 8.0,
                    41 : 2.0,
    }

    return convertor

def globcover_to_land_mask():
    """Generate a converter dictionary to be used by `lulc.lulc_to_x` to convert
    from GLOBCOVER-lulc to the ETlook `land_mask`.

    Returns
    -------
    dict
        Describes the conversions to be made.
    """

    convertor = {
                    11: 1,    
                    14:	1,    
                    20:	1,    
                    30:	1,    
                    40:	1,    
                    50:	1,    
                    60:	1,    
                    70:	1,    
                    90:	1,    
                    100: 1,   	 
                    110: 1,   	 
                    120: 1,   	 
                    130: 1,  	 
                    140: 1,  	 
                    150: 1,  	 
                    160: 1,  	 
                    170: 1,  	 
                    180: 1,  	 
                    190: 3,  	 
                    200: 1,  	 
                    210: 2,  	 
                    220: 1,  	 
                    230: 0  	 
        }

    return convertor

def globcover_to_lue_max():
    """Generate a converter dictionary to be used by `lulc.lulc_to_x` to convert
    from GLOBCOVER-lulc to the ETlook `lue_max`.

    Returns
    -------
    dict
        Describes the conversions to be made.
    """

    convertor = {
                11: 3.04,    
                14:	3.04,    
                20:	3.04,    
                30:	3.04,    
                40:	2.45,    
                50:	2.66,    
                60:	2.66,    
                70:	2.5,    
                90:	2.5,    
                100: 2.54,   	 
                110: 2.54,   	 
                120: 2.54,   	 
                130: 2.54,  	 
                140: 2.3,  	 
                150: 1.57,  	 
                160: 2.66,  	 
                170: 1.57,  	 
                180: 1.57,  	 
                190: 1.57,  	 
                200: 1.57,  	 
                210: 0,  	 
                220: 0,  	 
                230: 0  	 
        }

    return convertor

def globcover_to_rs_min():
    """Generate a converter dictionary to be used by `lulc.lulc_to_x` to convert
    from GLOBCOVER-lulc to the ETlook `rs_min`.

    Returns
    -------
    dict
        Describes the conversions to be made.
    """

    convertor = {
                11: 200,    
                14:	200,    
                20:	150,    
                30:	150,    
                40:	100,    
                50:	120,    
                60:	100,    
                70:	150,    
                90:	180,    
                100: 175,   	 
                110: 150,   	 
                120: 350,   	 
                130: 175,  	 
                140: 250,  	 
                150: 150,  	 
                160: 250,  	 
                170: 200,  	 
                180: 300,  	 
                190: 100,  	 
                200: 100,  	 
                210: 100,  	 
                220: 100,  	 
                230: 0  	 
        }

    return convertor

def globcover_to_z_obst_max():
    """Generate a converter dictionary to be used by `lulc.lulc_to_x` to convert
    from GLOBCOVER-lulc to the ETlook `z_obst_max`.

    Returns
    -------
    dict
        Describes the conversions to be made.
    """

    convertor = {
            11: 4.0,    
            14:	4.0,    
            20:	2.0,    
            30:	3.5,    
            40:	0.1,    
            50:	0.6,    
            60:	1.2,    
            70:	2.0,    
            90:	5.0,    
            100: 8.0,   	 
            110: 2.0,   	 
            120: 8.0,   	 
            130: 4.0,  	 
            140: 2.0,  	 
            150: 1.0,  	 
            160: 0.3,  	 
            170: 6.0,  	 
            180: 10,  	 
            190: 10,
            200: 0.1,
            210: 0.1,
            220: 0.1, 
            230: 0,
    }

    return convertor



import xarray as xr
import numpy as np
import os
import tqdm
import rasterio
import rioxarray.merge
import tempfile
from pydap.cas.urs import setup_session
import urllib.parse
from pywapor.general.logger import log
from rasterio.crs import CRS
from pywapor.general.processing_functions import save_ds, process_ds, remove_ds
import warnings
from pywapor.enhancers.apply_enhancers import apply_enhancer
from pywapor.collect.protocol.crawler import download_url, download_urls

def download(fp, product_name, coords, variables, post_processors, 
                fn_func, url_func, un_pw = None, tiles = None,  
                data_source_crs = None, parallel = False, spatial_tiles = True, 
                request_dims = True, timedelta = None):

    folder = os.path.split(fp)[0]

    # Create selection object.
    selection = create_selection(coords, target_crs = data_source_crs)

    # Make output filepaths, should be same length as `urls`.
    fps = [os.path.join(folder, fn_func(product_name, x)) for x in tiles]

    # Make data request URLs.
    session = start_session(url_func(product_name, tiles[0]), selection, un_pw)
    if spatial_tiles:
        idxss = [find_idxs(url_func(product_name, x), selection, session) for x in tqdm.tqdm(tiles, delay = 20)]
        urls = [create_url(url_func(product_name, x), idxs, variables, request_dims = request_dims) for x, idxs in zip(tiles, idxss)]
    else:
        idxs = find_idxs(url_func(product_name, tiles[0]), selection, session)
        urls = [create_url(url_func(product_name, x), idxs, variables, request_dims = request_dims) for x in tiles]

    # Download data.
    files = download_urls(urls, "", session, fps = fps, parallel = parallel)

    # Merge spatial tiles.
    if spatial_tiles:
        dss = [process_ds(xr.open_dataset(x, decode_coords = "all"), coords, variables, crs = data_source_crs) for x in files]
        ds = rioxarray.merge.merge_datasets(dss)
    else:
        dss = files
        ds = process_ds(xr.open_mfdataset(files, decode_coords = "all"), coords, variables, crs = data_source_crs)

    # Reproject if necessary.
    if ds.rio.crs.to_epsg() != 4326:
        ds = ds.rio.reproject(rasterio.crs.CRS.from_epsg(4326))

    # Apply product specific functions.
    for var, funcs in post_processors.items():
        for func in funcs:
            ds, label = apply_enhancer(ds, var, func)
            log.info(label)

    if isinstance(timedelta, np.timedelta64):
        ds["time"] = ds["time"] + timedelta

    # Remove unrequested variables.
    ds = ds[list(post_processors.keys())]
    
    # Save final output.
    ds.attrs = {}
    ds = save_ds(ds, fp, encoding = "initiate", label = "Saving merged data.")

    # Remove temporary files.
    if not isinstance(dss, type(None)):
        for x in dss:
            remove_ds(x)

    return ds

def find_idxs(base_url, selection, session):
    def _find_idxs(ds, k, search_range):
        all_idxs = np.where((ds[k] >= search_range[0]) & (ds[k] <= search_range[1]))[0]
        return [np.min(all_idxs), np.max(all_idxs)]
    fp = tempfile.NamedTemporaryFile(suffix=".nc").name
    url_coords = base_url + urllib.parse.quote(",".join(selection.keys()))
    fp = download_url(url_coords, fp, session, waitbar = False)
    ds = xr.open_dataset(fp, decode_coords = "all")
    idxs = {k: _find_idxs(ds, k, v) for k, v in selection.items()}
    return idxs

def create_url(base_url, idxs, variables, request_dims = True):
    if request_dims:
        dims = [f"{k}[{v[0]}:{v[1]}]" for k, v in idxs.items()]
    else:
        dims = []
    varis = [f"{k}{''.join([f'[{idxs[dim][0]}:{idxs[dim][1]}]' for dim in v[0]])}" for k, v in variables.items()]
    url = base_url + urllib.parse.quote(",".join(dims + varis))
    return url

def start_session(base_url, selection, un_pw = [None, None]):
    if un_pw == [None, None]:
        warnings.filterwarnings("ignore", "password was not set. ")
    url_coords = base_url + urllib.parse.quote(",".join(selection.keys()))
    session = setup_session(*un_pw, check_url = url_coords)
    return session

def download_xarray(url, fp, coords, variables, post_processors, 
                    data_source_crs = None, timedelta = None):

    warnings.filterwarnings("ignore", category=xr.SerializationWarning)
    online_ds = xr.open_dataset(url, decode_coords="all")
    # warnings.filterwarnings("default", category=xr.SerializationWarning)

    # Define selection.
    selection = create_selection(coords, target_crs = data_source_crs)

    # Make the selection on the remote.
    online_ds = online_ds.sel({k: slice(*v) for k, v in selection.items()})

    # Rename variables and assign crs.
    online_ds = process_ds(online_ds, coords, variables, crs = data_source_crs)

    # Download the data.
    ds = save_ds(online_ds, fp.replace(".nc", "_temp.nc"), label = f"Downloading data.")

    # Apply product specific functions.
    for var, funcs in post_processors.items():
        for func in funcs:
            ds, label = apply_enhancer(ds, var, func)
            log.info(label)

    if isinstance(timedelta, np.timedelta64):
        ds["time"] = ds["time"] + timedelta

    # Save final output
    out = save_ds(ds, fp, encoding = "initiate", label = "Saving netCDF.")

    remove_ds(ds)

    return out

def create_selection(coords, ds = None, target_crs = None, 
                        source_crs = CRS.from_epsg(4326)):
    """Create a dictionary that can be given to `xr.Dataset.sel`.

    Parameters
    ----------
    coords : dict
        Dictionary describing the different dimensions over which to select. Possible keys
        are "x" for latitude, "y" for longitude and "t" for time, but other selections
        keys are also allowed (e.g. so select a band). Values are tuples with the first
        value the respective dimension names in the `ds` and the second value the selector.
    ds : xr.Dataset, optional
        Dataset on which the selection will be applied, can be supplied to check if 
        dimensions are sorted ascending or descending and to derive `target_crs`, by default None.
    target_crs : rasterio.crs.CRS, optional
        crs of the dataset on which the selection will be applied, by default None.
    source_crs : rasterio.crs.CRS, optional
        crs of the `x` and `y` limits in `coords`, by default `epsg:4326`.

    Returns
    -------
    dict
        Dimension names with slices to apply to each dimension.
    """
    selection = {}

    if not isinstance(target_crs, type(None)):
        coords["x"][1], coords["y"][1] = rasterio.warp.transform(source_crs,
                                                target_crs,
                                                coords["x"][1], coords["y"][1])
    elif not isinstance(ds, type(None)):
        if ds.rio.crs != source_crs:
            coords["x"][1], coords["y"][1] = rasterio.warp.transform(source_crs,
                                        ds.rio.crs,
                                        coords["x"][1], coords["y"][1]) 

    if "t" in coords.keys():
        coords["t"][1] = [np.datetime64(t) for t in coords["t"][1]]

    for name, lim in coords.values():
        # If `ds` is given, checks if the dimensions are sorted from high to low
        # or from low to high and adjusts the limit accordingly.
        if isinstance(ds, xr.Dataset) and hasattr(lim, "__iter__"):
            lim = {False: lim, True: lim[::-1]}[bool(ds[name][0] - ds[name][1] > 0)]
        # Create a slice object if `lim` is a list or tuple.
        # if hasattr(lim, "__iter__"):
        #     lim = slice(*lim)
        selection[name] = lim

    return selection


from pywapor.general.logger import log, adjust_logger
import types
import functools
import numpy as np

def collect_sources(folder, sources, latlim, lonlim, timelim, return_fps = True):
    
    reversed_sources, reversed_enhancers = reverse_sources(sources)

    dss = dict()

    max_attempts = 2
    attempts = {k: 0 for k in reversed_sources.keys()}

    while not np.all([v >= max_attempts for v in attempts.values()]):

        reversed_sources = {k: v for k, v in reversed_sources.items() if attempts[k] < max_attempts}

        for (source, product_name), req_vars in reversed_sources.items():
            
            if isinstance(source, str):
                dl_module = __import__(f"pywapor.collect.product.{source}", 
                                    fromlist=[source])
                dler = dl_module.download
                source_name = source
            elif isinstance(source, types.FunctionType):
                dler = source
                source_name = source.__name__
            elif isinstance(source, functools.partial):
                dler = source
                source_name = source.func.__name__

            log.info(f"--> Collecting `{'`, `'.join(req_vars)}` from `{source_name}.{product_name}`.")
            log.push().add()
            
            args = {
                "folder": folder,
                "latlim": latlim,
                "lonlim": lonlim,
                "timelim": timelim,
                "product_name": product_name,
                "req_vars": req_vars,
                "post_processors": reversed_enhancers[(source, product_name)]
            }

            try:
                x = dler(**args)
                attempts[(source, product_name)] += max_attempts * 10
            except Exception as e:

                if "NetCDF: Filter error: unimplemented filter encountered" in e.args[0]:
                    info_url = r"https://github.com/Unidata/netcdf4-python/issues/1182"
                    log.warning(f"--> Looks like you installed `netcdf4` incorrectly, see {info_url} for more info.")

                if attempts[(source, product_name)] < max_attempts - 1:
                    log.warning(f"--> Collect attempt {attempts[(source, product_name)] + 1} of {max_attempts} for `{source_name}.{product_name}` failed, trying again after other sources have been collected. ({type(e).__name__}: {e}).")
                else:
                    log.warning(f"--> Collect attempt {attempts[(source, product_name)] + 1} of {max_attempts} for `{source_name}.{product_name}` failed, giving up now, see full traceback below for more info. ({type(e).__name__}: {e}).")
                    log.exception("")

                attempts[(source, product_name)] += 1
            else:
                if "time" in x.coords:
                    stime = np.datetime_as_string(x.time.values[0], unit = "m")
                    etime = np.datetime_as_string(x.time.values[-1], unit = "m")
                    log.add().info(f"> timesize: {x.time.size} [{stime}, ..., {etime}]").sub()
                dss[(source_name, product_name)] = x
            finally:
                log.pop()
                continue

    if return_fps:
        for key, ds in dss.items():
            fp = ds.encoding["source"]
            ds = ds.close()
            dss[key] = fp

    reversed_sources = {k: v for k, v in reversed_sources.items() if attempts[k] <= max_attempts}
    sources = trim_sources(reversed_sources, sources)

    return dss, sources

def reverse_sources(sources):
    reversed_sources = dict()
    reversed_enhancers = dict()
    for var, value in sources.items():
        for src in value["products"]:
            key = (src["source"], src["product_name"])
            enhancers = src["enhancers"]

            if key in reversed_sources.keys():
                reversed_sources[key].append(var)
                reversed_enhancers[key][var] = enhancers
            else:
                reversed_sources[key] = [var]
                reversed_enhancers[key] = {var: enhancers}

    return reversed_sources, reversed_enhancers

def trim_sources(reversed_sources, sources):
    for product, varis in reversed_sources.items():
        for var in varis:
            products_for_var = sources[var]["products"]
            idxs = [i for i, prod in enumerate(products_for_var) if (prod["source"], prod["product_name"]) == product]
            if len(idxs) == 0:
                continue
            elif len(idxs) == 1:
                removed_prod = sources[var]["products"].pop(idxs[0])
                log.warning(f"--> Continuing run without `{var}` from `{removed_prod['product_name']}`.")
            else:
                raise ValueError("Multiple identical `source.product`s for one var cannot exist")
        for var in varis:
            products_for_var = sources[var]["products"]
            if len(products_for_var) == 0:
                _ = sources.pop(var)
                log.warning(f"--> Continuing run without `{var}`, couldn't collect any product.")
    return sources

if __name__ == "__main__":

    return_fps = False

#     import datetime

#     sources = {

#         "r0":           {"products": [{"source": "MODIS", "product_name": "MCD43A3.061", "enhancers": "default"}]},
#         "z":            {"products": [{"source": "SRTM", "product_name": "30M", "enhancers": "default"}]},
#         "p":            {"products": [{"source": "CHIRPS", "product_name": "P05", "enhancers": "default"}]},
#         "lw_offset":    {"products": [{"source": "STATICS", "product_name": "WaPOR2", "enhancers": "default"}]},
#     }

#     folder = r"/Users/hmcoerver/Local/supersafe_dler"
#     latlim = [28.9, 29.7]
#     lonlim = [30.2, 31.2]
#     timelim = [datetime.date(2020, 7, 1), datetime.date(2020, 7, 15)]

#     adjust_logger(True, folder, "INFO")

#     dss0 = collect_sources(folder, sources, latlim, lonlim, timelim)



"""
https://opendap.cr.usgs.gov/opendap/hyrax/SRTMGL1_NUMNC.003/contents.html
"""
import datetime
import os
import json
import pywapor.collect
import xarray as xr
from pywapor.general.processing_functions import open_ds, remove_ds, save_ds
import pywapor.collect.accounts as accounts
from shapely.geometry.polygon import Polygon
from shapely.geometry import shape
import pywapor.collect.protocol.opendap as opendap
from pywapor.enhancers.dem import calc_slope, calc_aspect
import numpy as np

def tiles_intersect(latlim, lonlim):
    with open(os.path.join(pywapor.collect.__path__[0], "product/SRTM30_tiles.geojson")) as f:
        features = json.load(f)["features"]
    aoi = Polygon.from_bounds(lonlim[0], latlim[0], lonlim[1], latlim[1])
    tiles = list()
    for feature in features:
        shp = shape(feature["geometry"])
        tile = feature["properties"]["dataFile"]
        if shp.intersects(aoi):
            tiles.append(tile.split(".")[0])
    return tiles

def default_vars(product_name, req_vars = ["z"]):

    variables = {
        "30M": {
            "SRTMGL1_DEM": [("time", "lat", "lon"), "z"],
            "crs": [(), "spatial_ref"],
            }
    }

    req_dl_vars = {
        "30M": {
            "z": ["SRTMGL1_DEM", "crs"],
            "slope": ["SRTMGL1_DEM", "crs"],
            "aspect": ["SRTMGL1_DEM", "crs"],
        }
    }

    out = {val:variables[product_name][val] for sublist in map(req_dl_vars[product_name].get, req_vars) for val in sublist}
    
    return out

def drop_time(ds):
    return ds.isel(time=0).drop("time")

def default_post_processors(product_name, req_vars = ["z"]):
    
    post_processors = {
        "30M": {
            "z": [],
            "aspect": [calc_aspect],
            "slope": [calc_slope],
        }
    }

    out = {k:v for k,v in post_processors[product_name].items() if k in req_vars}

    return out

def fn_func(product_name, tile):
    fn = f"{product_name}_{tile}.nc"
    return fn

def url_func(product_name, tile):
    url = f"https://opendap.cr.usgs.gov/opendap/hyrax/SRTMGL1_NC.003/{tile}.SRTMGL1_NC.ncml.nc4?"
    return url

def download(folder, latlim, lonlim, product_name = "30M", req_vars = ["z"], variables = None, post_processors = None, **kwargs):
    folder = os.path.join(folder, "SRTM")

    appending = False
    fn = os.path.join(folder, f"{product_name}.nc")
    if os.path.isfile(fn):
        os.rename(fn, fn.replace(".nc", "_to_be_appended.nc"))
        existing_ds = open_ds(fn.replace(".nc", "_to_be_appended.nc"))
        if np.all([x in existing_ds.data_vars for x in req_vars]):
            existing_ds = existing_ds.close()
            os.rename(fn.replace(".nc", "_to_be_appended.nc"), fn)
            existing_ds = open_ds(fn)
            return existing_ds[req_vars]
        else:
            appending = True
            fn = os.path.join(folder, f"{product_name}_appendix.nc")
            req_vars = [x for x in req_vars if x not in existing_ds.data_vars]

    spatial_buffer = True
    if spatial_buffer:
        dx = dy = 0.0002777777777777768
        latlim = [latlim[0] - dy, latlim[1] + dy]
        lonlim = [lonlim[0] - dx, lonlim[1] + dx]

    timelim = [datetime.date(2000, 2, 10), datetime.date(2000, 2, 12)]
    tiles = tiles_intersect(latlim, lonlim)
        
    coords = {"x": ["lon", lonlim], "y": ["lat", latlim], "t": ["time", timelim]}
    
    if isinstance(variables, type(None)):
        variables = default_vars(product_name, req_vars)

    if isinstance(post_processors, type(None)):
        post_processors = default_post_processors(product_name, req_vars)
    else:
        default_processors = default_post_processors(product_name, req_vars)
        post_processors = {k: {True: default_processors[k], False: v}[v == "default"] for k,v in post_processors.items() if k in req_vars}

    data_source_crs = None
    parallel = False
    spatial_tiles = True
    un_pw = accounts.get("NASA")
    request_dims = True

    ds_new = opendap.download(fn, product_name, coords, 
                variables, post_processors, fn_func, url_func, un_pw = un_pw, 
                tiles = tiles, data_source_crs = data_source_crs, parallel = parallel, 
                spatial_tiles = spatial_tiles, request_dims = request_dims)

    if appending:
        ds = xr.merge([ds_new, existing_ds])
        lbl = f"Appending new variables (`{'`, `'.join(req_vars)}`) to existing file."
        ds = save_ds(ds, os.path.join(folder, f"{product_name}.nc"), encoding = "initiate", label = lbl)
        remove_ds(ds_new)
        remove_ds(existing_ds)
    else:
        ds = ds_new

    return ds

if __name__ == "__main__":

    folder = r"/Users/hmcoerver/Local/cog2_test"
    # latlim = [26.9, 33.7]
    # lonlim = [25.2, 37.2]
    latlim = [28.9, 29.7]
    lonlim = [30.2, 31.2]
    timelim = [datetime.date(2020, 7, 1), datetime.date(2020, 7, 11)]
    req_vars = ["z", "aspect"]

    # fn = os.path.join(os.path.join(folder, "SRTM"), "30M.nc")
    # if os.path.isfile(fn):
    #     os.remove(fn)

    # # SRTM.
    ds = download(folder, latlim, lonlim, req_vars = req_vars)
    print(ds.rio.crs, ds.rio.grid_mapping)




import os
from pywapor.collect.protocol import cog
from pywapor.general.processing_functions import open_ds, remove_ds, save_ds
import xarray as xr
import numpy as np

def default_vars(product_name, req_vars):
    variables = {
        'WaPOR2': {
                "Band1": [("lat", "lon"), "z_oro"],
                "Band2": [("lat", "lon"), "vpd_slope"],
                "Band3": [("lat", "lon"), "t_amp_year"],
                "Band4": [("lat", "lon"), "rn_offset"],
                "Band5": [("lat", "lon"), "r0_full"],
                "Band6": [("lat", "lon"), "r0_bare"],
                "Band7": [("lat", "lon"), "rs_min"],
                "Band8": [("lat", "lon"), "rn_slope"],
                "Band9": [("lat", "lon"), "t_opt"],
                "Band10": [("lat", "lon"), "lw_slope"],
                "Band11": [("lat", "lon"), "land_mask"],
                "Band12": [("lat", "lon"), "lw_offset"],
                "Band13": [("lat", "lon"), "z_obst_max"],
                "crs": [(), "spatial_ref"],
                    },
        'WaPOR3': {
                "Band2": [("lat", "lon"), "lw_offset"],
                "Band1": [("lat", "lon"), "lw_slope"],
                "crs": [(), "spatial_ref"],
        }
    }

    req_dl_vars = {
        "WaPOR2": {
            "z_oro": ["Band1", "crs"],
            "vpd_slope": ["Band2", "crs"],
            "t_amp_year": ["Band3", "crs"],
            "rn_offset": ["Band4", "crs"],
            "r0_full": ["Band5", "crs"],
            "r0_bare": ["Band6", "crs"],
            "rs_min": ["Band7", "crs"],
            "rn_slope": ["Band8", "crs"],
            "t_opt": ["Band9", "crs"],
            "lw_slope": ["Band10", "crs"],
            "land_mask": ["Band11", "crs"],
            "lw_offset": ["Band12", "crs"],
            "z_obst_max": ["Band13", "crs"],
        },
        "WaPOR3": {
            "lw_offset": ["Band2", "crs"],
            "lw_slope": ["Band1", "crs"],
        },
    }

    out = {val:variables[product_name][val] for sublist in map(req_dl_vars[product_name].get, req_vars) for val in sublist}
    
    return out

def scale_factor(ds, var, scale = 0.01):
    ds[var] = ds[var] * scale
    return ds

def default_post_processors(product_name, req_vars):
    post_processors = {
        'WaPOR2': {
                    "land_mask": [],
                    "lw_offset": [],
                    "lw_slope": [],
                    "r0_bare": [],
                    "r0_full": [],
                    "rn_offset": [],
                    "rn_slope": [],
                    "rs_min": [],
                    "t_amp_year": [],
                    "t_opt": [],
                    "vpd_slope": [],
                    "z_obst_max": [],
                    "z_oro": [],
                    },
        'WaPOR3': {
                    "lw_offset": [],
                    "lw_slope": [],
                    },
    }

    out = {k:v for k,v in post_processors[product_name].items() if k in req_vars}

    return out

def url_func(product_name):
    url = f"https://storage.googleapis.com/fao-cog-data/{product_name}.tif"
    return url

def download(folder, latlim, lonlim, product_name, req_vars,
                variables = None, post_processors = None, **kwargs):
    
    folder = os.path.join(folder, "STATICS")
    
    appending = False
    fn = os.path.join(folder, f"{product_name}.nc")
    if os.path.isfile(fn):
        os.rename(fn, fn.replace(".nc", "_to_be_appended.nc"))
        existing_ds = open_ds(fn.replace(".nc", "_to_be_appended.nc"))
        if np.all([x in existing_ds.data_vars for x in req_vars]):
            existing_ds = existing_ds.close()
            os.rename(fn.replace(".nc", "_to_be_appended.nc"), fn)
            existing_ds = open_ds(fn)
            return existing_ds[req_vars]
        else:
            appending = True
            fn = os.path.join(folder, f"{product_name}_appendix.nc")
            req_vars = [x for x in req_vars if x not in existing_ds.data_vars]

    spatial_buffer = True
    if spatial_buffer:
        dx = dy = 0.05
        latlim = [latlim[0] - dy, latlim[1] + dy]
        lonlim = [lonlim[0] - dx, lonlim[1] + dx]

    coords = {"x": ("lon", lonlim), "y": ("lat", latlim)}

    if isinstance(variables, type(None)):
        variables = default_vars(product_name, req_vars)

    if isinstance(post_processors, type(None)):
        post_processors = default_post_processors(product_name, req_vars)
    else:
        default_processors = default_post_processors(product_name, req_vars)
        post_processors = {k: {True: default_processors[k], False: v}[v == "default"] for k,v in post_processors.items() if k in req_vars}

    ds_new = cog.download(fn, product_name, coords, variables, 
                        post_processors, url_func)

    if appending:
        ds = xr.merge([ds_new, existing_ds])
        lbl = f"Appending new variables (`{'`, `'.join(req_vars)}`) to existing file."
        ds = save_ds(ds, os.path.join(folder, f"{product_name}.nc"), encoding = "initiate", label = lbl)
        remove_ds(ds_new)
        remove_ds(existing_ds)
    else:
        ds = ds_new

    return ds

if __name__ == "__main__":

    req_vars = [
                'land_mask',
                # 'lw_offset',
                'lw_slope',
                # 'r0_bare',
                'r0_full',
                'rn_offset',
                'rn_slope',
                'rs_min',
                't_amp_year',
                't_opt',
                # 'vpd_slope',
                'z_obst_max',
                'z_oro'
                ]

    product_name = "WaPOR2"

    folder = r"/Users/hmcoerver/Local/statics_test4"
    latlim = [28.9, 29.7]
    lonlim = [30.2, 31.2]

    ds = download(folder, latlim, lonlim, product_name, req_vars,
                variables = None, post_processors = None)



# -*- coding: utf-8 -*-
from pywapor.general import variables, processing_functions, compositer, pre_defaults, lazifier, network, reproject, levels

__all__ = ['variables', 'processing_functions', 'compositer', 'pre_defaults', 'lazifier', 'network', 'reproject', 'levels']

__version__ = '0.1'



import numpy as np

def adjust_wind_height(ds, var, z = 10):
    """
    See FAO56 equation 47.
    """
    if "2m" in var:
        old_name = var.replace("2m", f"{z}m")
    else:
        old_name = var
    ds[var] = ds[old_name] * ((4.87)/(np.log(67.8*z - 5.42)))
    return ds

def windspeed(ds, var):
    if np.all([x in ds.data_vars for x in ["u10m", "v10m"]]):
        ds[var] = np.hypot(ds["u10m"], ds["v10m"])
    return ds




"""Importing `log` from this module starts a logging handler, which can be logged to 
with `log.info("log this info")` etc.
"""
from python_log_indenter import IndentedLoggerAdapter
import logging
import os
from tqdm.contrib.logging import logging_redirect_tqdm

log_settings = logging.getLogger(__name__)
__formatter__ = logging.Formatter(fmt='%(message)s', datefmt = '%Y/%m/%d %H:%M:%S')
__handler__ = logging.StreamHandler()
__handler__.setFormatter(__formatter__)
__handler__.setLevel("INFO")
log_settings.addHandler(__handler__)
log = IndentedLoggerAdapter(log_settings)
# logging_redirect_tqdm(loggers = [log_settings])

def adjust_logger(log_write, folder, log_level):
    """Function to adjust the default logging settings that get initiated by
    importing this module.

    Parameters
    ----------
    log_write : bool
        Stop or start writing to log.
    folder : str
        Path to folder in which to store `"log.txt"`.
    log_level : str
        Set the log level.
    """
    if log_write and not any([isinstance(x, logging.FileHandler) for x in log_settings.handlers]):
        if not os.path.isdir(folder):
            os.makedirs(folder)
        handler = logging.FileHandler(filename = os.path.join(folder, "log.txt"))
        formatter = logging.Formatter('{asctime} {levelname:>9s}: {message}', style='{')
        handler.setFormatter(formatter)
        handler.setLevel("DEBUG")
        log_settings.addHandler(handler)
        logging_redirect_tqdm(loggers = log_settings.handlers)
    log_settings.setLevel(log_level)


import xarray as xr
import numpy as np
import matplotlib.pyplot as plt
from scipy.spatial import Delaunay, cKDTree
from dask.diagnostics import ProgressBar

def create_grid(input_ds, dx, dy, bb = None, precision = 0.1):

    # precision: The left and lower bounds will snap to a `1/precision`-th degree 
    # below smallest coordinate in input_ds.

    # Open input dataset if necessary.
    if isinstance(input_ds, str):
        input_ds = xr.open_dataset(input_ds, chunks = "auto")

    if isinstance(bb, type(None)):
        bb = [input_ds.x.min().values, input_ds.y.min().values, input_ds.x.max().values, input_ds.y.max().values]

    # Determine box that is just larger then domain in input_ds.
    xmin = np.floor(bb[0] * precision**-1) / precision**-1
    xmax = np.ceil(bb[2] * precision**-1) / precision**-1
    ymin = np.floor(bb[1] * precision**-1) / precision**-1
    ymax = np.ceil(bb[3] * precision**-1) / precision**-1

    # Determine coordinates.
    gridx = np.arange(xmin, xmax + dx, dx)
    gridy = np.arange(ymin, ymax + dy, dy)

    # Wrap coordinates in xr.Dataset.
    grid_ds = xr.Dataset(None, coords = {"y": gridy, "x": gridx})
    
    # Add relevant variables.
    for var in input_ds.data_vars:
        dims = [(x, input_ds[x]) for x in input_ds[var].dims if x not in input_ds.y.dims]
        dims += [("y", grid_ds.y), ("x", grid_ds.x)]
        grid_ds[var] = xr.DataArray(coords = {k:v for k,v, in dims}, dims = [x[0] for x in dims])

    return grid_ds

def regrid(grid_ds, input_ds, max_px_dist = 10):

    # Create output dataset
    output_ds = grid_ds.stack({"grid_pixel": ("y", "x")})
    no_pixel = output_ds.grid_pixel.size

    # Create intermediate dataset without multi-index
    grid_adj_ds = output_ds.assign_coords({"i": ("grid_pixel", range(no_pixel))}).set_index(grid_pixel="i")
    grid_adj_ds = grid_adj_ds.drop(["x", "y"]).assign({"x": ("grid_pixel", grid_adj_ds.x.values),
                                                       "y": ("grid_pixel", grid_adj_ds.y.values)})

    # Determine bounding box of current chunk.
    bb = [grid_adj_ds.x.min(), grid_adj_ds.y.min(), grid_adj_ds.x.max(), grid_adj_ds.y.max()]

    # Determine pixel size.
    dx = np.median(np.unique(np.diff(grid_ds.x)))
    dy = np.median(np.unique(np.diff(grid_ds.y)))

    # Open input dataset if necessary.
    if isinstance(input_ds, str):
        input_ds = xr.open_dataset(input_ds, chunks = "auto")

    # Move `x` and `y` to variables, so they can be masked later on.
    input_ds = input_ds.reset_coords(["x", "y"])

    # Filter out irrelevant input data for current chunk.
    mask = ((input_ds.y >= bb[1] - dy) & 
            (input_ds.y <= bb[3] + dy) & 
            (input_ds.x >= bb[0] - dx) & 
            (input_ds.x <= bb[2] + dx))

    # Mask irrelevant pixels.
    data = input_ds.where(mask, drop = False)

    # Check if there is enough input data for current chunk.
    if np.any(np.array([data[x].count().values for x in data.data_vars if x not in ["x", "y"]]) < 10):
        return output_ds.unstack()

    # --heavy-> Transform input data from 2D to 1D and remove empty pixels.
    data_pixel = data.stack({"pixel": input_ds.x.dims}).dropna("pixel")

    # Load input data coordinates for scipy.
    xy = np.dstack((data_pixel.x.values,
                    data_pixel.y.values))[0]

    # Determine distances between grid pixel and nearest input data point.
    tree = cKDTree(xy)
    xi = np.stack(np.meshgrid(grid_ds.x, grid_ds.y), axis = 2)
    dists = xr.DataArray(tree.query(xi)[0], dims = ("y", "x"), coords = { "y": grid_ds.y, "x": grid_ds.x})
    dists_pixel = dists.stack({"grid_pixel": ("y", "x")}).drop("grid_pixel")

    # Load pixel coordinates for scipy functions, excluding pixels 
    # for which input data is too far away.
    max_dist = np.min([max_px_dist * dx, max_px_dist * dy])
    grid_adj_ds = grid_adj_ds.where(dists_pixel < max_dist, drop = True)
    uv = np.dstack((grid_adj_ds.x.values, grid_adj_ds.y.values))[0]

    # 2D Delaunay tessellation.
    # Also see: https://stackoverflow.com/questions/20915502/speedup-scipy-griddata-for-multiple-interpolations-between-two-irregular-grids
    tri = Delaunay(xy)

    # Find simplex for grid points.
    simplex = tri.find_simplex(uv)

    # Determine vertices and weights.
    vtx = np.take(tri.simplices, simplex, axis = 0)
    temp = np.take(tri.transform, simplex, axis = 0)
    delta = uv - temp[:, 2]
    bary = np.einsum('njk,nk->nj', temp[:, :2, :], delta)
    wts = np.hstack((bary, 1 - bary.sum(axis = 1, keepdims = True)))

    # Wrap vertices and weights in xr.DataArray.
    kwargs = {"dims": ("grid_pixel", "j"), 
              "coords": {"j": range(3), "grid_pixel": grid_adj_ds.grid_pixel}}
    wts_da = xr.DataArray(wts, **kwargs)
    vtx_da = xr.DataArray(vtx, **kwargs)

    # --heavy-> Select relevant input data.
    vls = data_pixel.drop(["x", "y"]).isel(pixel = vtx_da)

    # Apply weights to input data.
    for var in vls.data_vars:
        da = xr.dot(vls[var], wts_da, dims = "j").where((wts_da > 0).all("j"))
        output_ds[var] = da.reindex(grid_pixel = dists_pixel.grid_pixel).drop("grid_pixel")

    return output_ds.unstack()

if __name__ == "__main__":

    ...
    # Small test.
    # input_ds = xr.tutorial.open_dataset("rasm")
    # input_ds = input_ds.rename_dims({"x": "nx", "y": "ny"}).rename_vars({"xc":"x", "yc": "y"})

    # grid_ds = create_grid(input_ds, 1.0, 1.0).chunk({"x":500, "y":500})

    # with ProgressBar():
    #     out = xr.map_blocks(regrid, grid_ds, (input_ds,), template = grid_ds).compute()

    # from cartopy import crs as ccrs

    # fig = plt.figure(figsize=(15, 10))
    # ax1 = plt.subplot(2, 1, 1, projection = ccrs.PlateCarree())
    # ax2 = plt.subplot(2, 1, 2, projection = ccrs.PlateCarree())
    # kwargs = {"x": "x", "y": "y", "vmin": -20, "vmax": 20}
    # bla = input_ds.isel(time=0).Tair.plot.pcolormesh(ax = ax1, **kwargs)
    # ax1.coastlines()
    # ax1.set_facecolor("lightgray")
    # ax1.gridlines(draw_labels = True, color='gray', linestyle=':')
    # ax1.set_title("curvilinear")
    # out.isel(time=0).Tair.plot.pcolormesh(ax = ax2, **kwargs)
    # ax2.coastlines()
    # ax2.set_facecolor("lightgray")
    # ax2.gridlines(draw_labels = True, color='gray', linestyle=':')
    # ax2.set_title("rectolinear")



from pywapor.et_look_v2_v3 import constants as c
import numpy as np
import xarray as xr

def interception_mm(P_24, vc, lai, int_max=0.2):
    r"""
    Computes the daily interception. The daily interception of a vegetated area
    is calculated according to :footcite:t:`braden1985energiehaushalts` 
    and :footcite:t:`hoyningen1983interception`.

    .. math ::

        I^*=I_{max} \cdot I_{lai} \cdot \left(1-\frac{1}{1+\frac{c_{veg} \cdot P}
        {I_{max} \cdot I_{lai}}}\right)

    Parameters
    ----------
    P_24 :  float
        daily rainfall, 
        :math:`P`
        [mm day :math:`^{-1}`]

    vc : float
        vegetation cover, 
        :math:`c_{veg}`
        [-]

    lai : float
        leaf area index, 
        :math:`I_{lai}`
        [-]

    int_max :  float
        maximum interception per leaf, 
        :math:`I_{max}`
        [mm day :math:`^{-1}`]

    Returns
    -------
    int_mm : float
        interception, 
        :math:`I^*`
        [mm day :math:`^{-1}`]

    """
    if isinstance(lai, xr.DataArray):
        zero_mask = np.logical_or(np.logical_or(vc == 0, lai ==0), P_24 == 0)
    else:
        zero_mask = np.logical_or.reduce((lai == 0, vc == 0, P_24 == 0))

    res = int_max * lai * (1 - (1 / (1 + ((vc * P_24) / (int_max * lai)))))

    if isinstance(res, xr.DataArray):
        res = xr.where(zero_mask, 0.0, res)
    else:
        res[zero_mask] = 0

    return res


def et_reference(rn_24_grass, ad_24, psy_24, vpd_24, ssvp_24, u_24):
    r"""
    Computes the reference evapotranspiration. 
    
    The reference evapotranspiration :math:`ET_{ref}` is an important concept 
    in irrigation science. The reference evapotranspiration can be inferred 
    from routine meteorological measurements. 
    
    The reference evapotranspiration is the evapotranspiration of grass under 
    well watered conditions. First the aerodynamical resistance for 
    grass :math:`r_{a,grass}` [sm :math:`^{-1}`] is calculated:

    .. math ::

        r_{a,grass}=\frac{208}{u_{obs}}

    Then the reference evapotranspiration :math:`ET_{ref}` [W m :math:`^{-2}`] can be calculated
    as follows, with taking the default value for the grass surface resistance
    :math:`r_{grass}` = 70 sm :math:`^{-1}`

    .. math ::

        ET_{ref}=\frac{\Delta \cdot Q_{grass}^{*}+
        \rho c_{p}\frac{\Delta_{e}}{r_{a,grass}}}
        {\Delta+\gamma \cdot \left(1+\frac{r_{grass}}{r_{a,grass}}\right)}

    The soil heat flux is assumed to be zero or close to zero on a daily basis.

    Parameters
    ----------
    rn_24_grass : float
        net radiation for reference grass surface, 
        :math:`Q^{*}_{grass}`
        [Wm-2]
    u_24 : float
        daily wind speed at observation height, 
        :math:`u_{obs}`
        [m/s]
    ad_24 : float
        daily air density, 
        :math:`\rho_{24}`
        [kg m-3]
    psy_24 : float
        daily psychrometric constant, 
        :math:`\gamma_{24}`
        [mbar K-1]
    vpd_24 : float
        daily vapour pressure deficit, 
        :math:`\Delta_{e,24}`
        [mbar]
    ssvp_24 : float
        daily slope of saturated vapour pressure curve, 
        :math:`\Delta_{24}`
        [mbar K-1]

    Returns
    -------
    et_ref_24 : float
        reference evapotranspiration (well watered grass) energy equivalent, 
        :math:`ET_{ref}`
        [W m-2]
    """
    r_grass = 70
    ra_grass = 208. / u_24
    et_ref_24 = (ssvp_24 * rn_24_grass + ad_24 * c.sh * (vpd_24 / ra_grass)) /\
        (ssvp_24 + psy_24 * (1 + r_grass / ra_grass))
    return et_ref_24


def et_reference_mm(et_ref_24, lh_24):
    r"""
    Computes the reference evapotranspiration.

    .. math ::

        ET_{ref}=ET_{ref} \cdot d_{sec} \cdot \lambda_{24}

    where:

    * :math:`d_{sec}` seconds in the day = 86400 [s]

    Parameters
    ----------
    et_ref_24 : float
        daily reference evapotranspiration energy equivalent, 
        :math:`ET_{ref}`
        [W m-2]
    lh_24 : float
        daily latent heat of evaporation, 
        :math:`\lambda_{24}`
        [J/kg]

    Returns
    -------
    et_ref_24_mm : float
        reference evapotranspiration (well watered grass), 
        :math:`ET_{ref}`
        [mm d-1]
    """
    x = et_ref_24 * c.day_sec / lh_24

    if isinstance(x, xr.DataArray):
        et_ref_24_mm = x.clip(0, np.inf)
    else:
        et_ref_24_mm = np.clip(x, 0, np.inf)

    return et_ref_24_mm


def et_actual_mm(e_24_mm, t_24_mm):
    r"""
    Computes the actual evapotranspiration based on the separate calculations
    of evaporation and transpiration.

    .. math ::

        ET = E + T

    Parameters
    ----------
    e_24_mm : float
        daily evaporation in mm, 
        :math:`E`
        [mm d-1]
    t_24_mm : float
        daily transpiration in mm, 
        :math:`T`
        [mm d-1]

    Returns
    -------
    et_24_mm : float
        daily evapotranspiration in mm, 
        :math:`ET`
        [mm d-1]
    """
    return e_24_mm + t_24_mm



"""Module containing functions that generate some default settings for `pywapor.pre_et_look`
and `pywapor.pre_se_root`.
"""
from pywapor.enhancers.temperature import lapse_rate
from functools import partial
import pywapor.enhancers.dem as dem
from pywapor.enhancers.temperature import kelvin_to_celsius
import pywapor.enhancers.lulc as lulc
from functools import partial

def constants_defaults():

    c = {
        'nd_min': 0.125,
        'nd_max': 0.8,
        'vc_pow': 0.7,
        'vc_min': 0,
        'vc_max': 0.9677324224821418,
        'lai_pow': -0.45,
        'diffusion_slope': -1.33,
        'diffusion_intercept': 1.15,
        't_min': 0,
        't_max': 50,
        'vp_slope': 0.14,
        'vp_offset': 0.34,
        'int_max': 0.2,
        'tenacity': 1.5,
        'rcan_max': 1000000,
        'ndvi_obs_min': 0.25,
        'ndvi_obs_max': 0.75,
        'obs_fr': 0.25,
        'z_obs': 2,
        'z_b': 100,
        'c1': 1,
        'iter_h': 3,
        'r_soil_pow': -2.1,
        'r_soil_min': 800,
        'se_top': 0.5,
        'porosity': 0.4,
        'r0_grass': 0.23,
        'eps_a': 0.5,

        # se_root
        'z0m_full': 0.1,
        'z0m_bare': 0.001,
        'aod550_i': 0.01,
        'fraction_h_bare': 0.65,
        'fraction_h_full': 0.95,
        'disp_bare': 0.0,
        'disp_full': 0.667,
        'r0_bare_wet': 0.2,
        'IO': 1367.,

        # Biomass
        'dh_ap': 52750, 
        'd_s': 704.98, 
        'dh_dp': 211000,
        'ar_slo': 0.0, 
        'ar_int': 0.5,
        'fpar_slope': 1.257, 
        'fpar_offset': -0.161,
        'o2': 20.9,
        'co2_ref': 281,
        'gcgdm': 0.4,
        'phot_eff': 2.49,



    }

    return c


# def composite_enhancements_defaults():
#     """Returns a dictionary with the default functions to apply to the calculated
#     composites.

#     Returns
#     -------
#     dict
#         Keys are variable names, values are lists of functions.
#     """
#     composite_enhancements = {
#         "t_air_24":     [lapse_rate],
#         "t_air_min_24": [lapse_rate],
#         "t_air_max_24": [lapse_rate],
#         "z":            [partial(dem.to_slope, out_var = "slope"), 
#                         partial(dem.to_aspect, out_var = "aspect"),
#                         partial(dem.to_lat, out_var = "lat_deg"),
#                         partial(dem.to_lon, out_var = "lon_deg"),]
#     }
#     return composite_enhancements

# def source_enhancements_defaults():
#     """Returns a dictionary with the default functions to apply to source
#     data.

#     Returns
#     -------
#     dict
#         Keys are tuples like `("source", "variable")`, values are lists of functions.
#     """

#     def remove_var(ds, var):
#         return ds.drop_vars([var])

#     source_enhancements = {
#         ("MERRA2",  "t_air_24"):        [kelvin_to_celsius],#
#         ("MERRA2",  "t_air_min_24"):    [kelvin_to_celsius],#
#         ("MERRA2",  "t_air_max_24"):    [kelvin_to_celsius],#
#         ("GEOS5",   "t_air_24"):        [kelvin_to_celsius],#
#         ("GEOS5",   "t_air_min_24"):    [kelvin_to_celsius],#
#         ("GEOS5",   "t_air_max_24"):    [kelvin_to_celsius],#
#         ("GLOBCOVER", "lulc"):          [partial(lulc.lulc_to_x, out_var = "land_mask", 
#                                                     convertor = lulc.globcover_to_land_mask()),
#                                         partial(lulc.lulc_to_x, out_var = "rs_min", 
#                                                     convertor = lulc.globcover_to_rs_min()),
#                                         partial(lulc.lulc_to_x, out_var = "lue_max", 
#                                                     convertor = lulc.globcover_to_lue_max()),
#                                         partial(lulc.lulc_to_x, out_var = "z_obst_max", 
#                                                     convertor = lulc.globcover_to_z_obst_max()),
#                                         remove_var,
#                                         ],
#         ("WAPOR", "lulc"):              [partial(lulc.lulc_to_x, out_var = "land_mask", 
#                                                     convertor = lulc.wapor_to_land_mask()),
#                                         partial(lulc.lulc_to_x, out_var = "rs_min", 
#                                                     convertor = lulc.wapor_to_rs_min()),
#                                         partial(lulc.lulc_to_x, out_var = "lue_max", 
#                                                     convertor = lulc.wapor_to_lue_max()),
#                                         partial(lulc.lulc_to_x, out_var = "z_obst_max", 
#                                                     convertor = lulc.wapor_to_z_obst_max()),
#                                         remove_var,
#                                         ],
#     }
#     return source_enhancements

# def composite_defaults():
#     """Returns the default composite settings for each variable.

#     Returns
#     -------
#     dict
#         Dictionary in which keys are variable names and values are dictionaries
#         with `cmeta` settings.
#     """

#     cdefaults = {

#         'ndvi': {
#                     "composite_type": "mean",
#                     "temporal_interp": "linear",
#                     "spatial_interp": "nearest",
#                     "var_name": "ndvi",
#                     "var_unit": "-",
#                 },
#         'r0': {
#                     "composite_type": "mean",
#                     "temporal_interp": "linear",
#                     "spatial_interp": "nearest",
#                     "var_name": "r0",
#                     "var_unit": "-",
#                 },
#         'lulc': {
#                     "composite_type": 'mean',
#                     "temporal_interp": 'nearest',
#                     "spatial_interp": "nearest",
#                     "var_name": "lulc",
#                     "var_unit": "-",
#                 },
#         'z': {
#                     "composite_type": False,
#                     "temporal_interp": False,
#                     "spatial_interp": "bilinear",
#                     "var_name": "z",
#                     "var_unit": "m",
#                 },
#         'p': {
#                     "composite_type": "mean",
#                     "temporal_interp": "linear",
#                     "spatial_interp": "nearest",
#                     "var_name": "p_24",
#                     "var_unit": "mm/day",
#                 },
#         'se_root': {
#                     "composite_type": "max",
#                     "temporal_interp": False,
#                     "spatial_interp": "nearest",
#                     "var_name": "se_root",
#                     "var_unit": "-",
#                 },
#         'ra': {
#                     "composite_type": "mean",
#                     "temporal_interp": "linear",
#                     "spatial_interp": "bilinear",
#                     "var_name": "ra_24",
#                     "var_unit": "-",
#                 }
#         }

#     meteo = {
#             "composite_type": "mean",
#             "temporal_interp": "linear",
#             "spatial_interp": "bilinear",
#             }

#     meteo_names_units = {"t_air_max": "°C", "t_air_min":"°C", "t_air":"°C", 
#         "u2m":"m/s", "v2m":"m/s","qv":"kg/kg","p_air_0":"kPa"}

#     for name, unit in meteo_names_units.items():
#         cdefaults[name] = {**{"var_name": name, "var_unit": unit}, **meteo}

#     static = {
#             "composite_type": False,
#             "temporal_interp": False,
#             "spatial_interp": "bilinear",    
#     }

#     statics_names_units = {'land_mask': "-",
#                         'lw_offset': "-",
#                         'lw_slope': "-",
#                         'r0_bare': "-",
#                         'r0_full': "-",
#                         'rn_offset': "-",
#                         'rn_slope': "-",
#                         'rs_min': "sm-1",
#                         't_amp_year': "C", 
#                         't_opt': "C",
#                         'vpd_slope': "mbar-1", 
#                         'z_obst_max': "m",
#                         'z_oro': "m",
#     }

#     for name, unit in statics_names_units.items():
#         cdefaults[name] = {**{"var_name": name, "var_unit": unit}, **static}

#     return cdefaults


from pywapor.et_look_v2_v3 import constants as c
from numpy import log


def initial_soil_aerodynamic_resistance(u_24, z_obs=2):
    r"""
    Computes the aerodynamic resistance for soil without stability corrections
    :math:`r_{a,soil}^{0}`.

    .. math ::

        r_{a,soil}^{0}=\frac{\ln\left(\frac{z_{obs}}{z_{0,soil}}\right)
                       \cdot \ln\left(\frac{z_{obs}}{0.1z_{0,soil}}\right)}
                       {k^{2} \cdot u_{obs}}


    where the following constants are used

    * :math:`z_{0,soil}` soil roughness = 0.001 [m]
    * :math:`k` = karman constant = 0.41 [-]

    The factor 0.1 is the ratio between the surface roughness for momentum and
    heat.

    Parameters
    ----------
    u_24 : float
        daily wind speed at observation height,
        :math:`u_{obs}`
        [m/s]

    z_obs : float
        observation height,
        :math:`z_{obs}`
        [m]

    Returns
    -------
    ra_soil_init : float
        aerodynamic resistance without stability corrections, 
        :math:`r_{a,soil}^{0}`
        [s/m]

    """
    return (log(z_obs / c.z0_soil) * log(z_obs / (0.1 * c.z0_soil))) / (c.k**2 * u_24)


def initial_daily_evaporation(rn_24_soil, g0_24, ssvp_24, ad_24, vpd_24,
                              psy_24, r_soil, ra_soil_init):
    r"""
    Computes the soil evaporation based on the Penman Monteith equation
    adapted for soil.

    .. math ::

        E^{0}=\frac{\Delta\left(Q_{soil}^{*}-G\right)+\rho c_{p}
        \frac{\Delta_{e}}{r_{a,soil}}}{\Delta+
        \gamma\left(1+\frac{r_{soil}}{r_{a,soil}}\right)}

    where the following constants are used

    * :math:`c_{p}` specific heat for dry air = 1004 [J kg-1 K-1]
    * :math:`k` = karman constant = 0.41 [-]

    Parameters
    ----------
    rn_24_soil : float
        daily net radiation for soil, 
        :math:`Q_{soil}^{*}`
        [W m-2]
    g0_24 : float
        daily soil heat flux, 
        :math:`G`
        [W m-2]
    ssvp_24 : float
       daily slope of saturated vapour pressure curve
       :math:`\Delta`
       [mbar K-1]
    ad_24 : float
        daily air density, 
        :math:`\rho`
        [kg m-3]
    vpd_24 : float
        daily vapour pressure deficit, 
        :math:`\Delta_{e}`
        [mbar]
    psy_24 : float
        daily psychrometric constant, 
        :math:`\gamma`
        [mbar K-1]
    r_soil : float
        soil resistance, 
        :math:`r_{soil}`
        [m s-1]
    ra_soil_init : float
        initial soil aerodynamic resistance, 
        :math:`r_{a,soil}`
        [m s-1]

    Returns
    -------
    e_24_init : float
        initial estimate radiation equivalent daily evaporation, 
        :math:`E^{0}`
        [W m-2]

    """
    numerator = (ssvp_24 * (rn_24_soil - g0_24) +
                 ad_24 * c.sh * (vpd_24 / ra_soil_init))
    denominator = (ssvp_24 + psy_24 * (1 + r_soil / ra_soil_init))
    return numerator / denominator


def initial_daily_evaporation_mm(e_24_init, lh_24):
    r"""
    Computes the soil evaporation based on the Penman Monteith equation
    adapted for soil.

    .. math ::

        E^{0}=E^{0} \cdot d_{sec} \cdot \lambda_{24}

    where the following constants are used

    * :math:`d_{sec}` seconds in the day = 86400 [s]

    Parameters
    ----------
    e_24_init : float
        initial estimate daily evaporation, 
        :math:`E^{0}`
        [W m-2]
    lh_24 : float
        daily latent heat of evaporation, 
        :math:`\lambda_{24}`
        [J/kg]

    Returns
    -------
    e_24_init_mm : float
        initial estimate daily evaporation in mm, 
        :math:`E^{0}`
        [mm d-1]
    """
    return e_24_init * c.day_sec / lh_24


def initial_canopy_aerodynamic_resistance(u_24, z0m, z_obs=2):
    r"""
    Computes the aerodynamic resistance for a canopy soil without stability
    corrections :math:`r_{a,}^{0}`.

    .. math ::

        r_{a,canopy}^{0}=\frac{\ln\left(\frac{z_{obs}}{z_{0,m}}\right) \cdot \ln
        \left(\frac{z_{obs}}{0.1z_{0,m}}\right)}{k^{2} \cdot u_{obs}}

    where the following constants are used

    * :math:`k` = karman constant = 0.41 [-]

    The factor 0.1 is the ratio between the surface roughness for momentum and
    heat.

    Parameters
    ----------
    u_24 : float
        daily wind speed at observation height, 
        :math:`u_{obs}`
        [m/s]
    z0m : float
        roughness length, 
        :math:`z_{0,m}`
        [m]
    z_obs : float
        observation height, 
        :math:`z_{obs}`
        [m]

    Returns
    -------
    ra_canopy_init : float
        canopy resistance without stability corrections, 
        :math:`r_{a,canopy}^{0}`
        [s/m]
    """
    return (log(z_obs / z0m) * log(z_obs / (0.1 * z0m))) / (c.k**2 * u_24)


def initial_daily_transpiration(rn_24_canopy, ssvp_24, ad_24, vpd_24,
                                psy_24, r_canopy, ra_canopy_init):
    r"""
    Computes the soil evaporation based on the Penman Monteith equation adapted
    for soil.

    .. math ::

        T_{0}=\frac{\Delta\left(Q_{canopy}^{*}\right)
        +\rho c_{p}\frac{\Delta_{e}}
        {r_{a,canopy}}}{\Delta+
        \gamma\left(1+\frac{r_{canopy}}{r_{a,canopy}}\right)}

    where the following constants are used

    * :math:`c_{p}` specific heat for dry air = 1004 [J kg-1 K-1]
    * :math:`k` = karman constant = 0.41 [-]

    Parameters
    ----------
    rn_24_canopy : float
        daily net radiation for the canopy, 
        :math:`Q_{soil}^{*}`
        [W m-2]
    ssvp_24 : float
       daily slope of saturated vapour pressure curve
       :math:`\Delta`
       [mbar K-1]
    ad_24 : float
        daily air density, 
        :math:`\rho`
        [kg m-3]
    vpd_24 : float
        daily vapour pressure deficit, 
        :math:`\Delta_{e}`
        [mbar]
    psy_24 : float
        daily psychrometric constant, 
        :math:`\gamma`
        [mbar K-1]
    r_canopy : float
        canopy resistance, 
        :math:`r_{canopy}`
        [m s-1]
    ra_canopy_init : float
        initial canopy aerodynamic resistance, 
        :math:`r_{a,canopy}`
        [m s-1]

    Returns
    -------
    t_24_init : float
        initial estimate radiation equivalent daily transpiration, 
        :math:`T^{0}`
        [W m-2]
    """
    numerator = (ssvp_24 * rn_24_canopy + ad_24 *
                 c.sh * (vpd_24 / ra_canopy_init))
    denominator = (ssvp_24 + psy_24 * (1 + r_canopy / ra_canopy_init))
    return numerator / denominator


def initial_daily_transpiration_mm(t_24_init, lh_24):
    r"""
    Computes the canopy transpiration based on the Penman Monteith equation
    adapted for canopy.

    .. math ::

        T^{0}=T^{0} \cdot d_{sec} \cdot \lambda_{24}

    where the following constants are used

    * :math:`d_{sec}` seconds in the day = 86400 [s]

    Parameters
    ----------
    t_24_init : float
        initial estimate daily transpiration, 
        :math:`E^{0}`
        [W m-2]
    lh_24 : float
        daily latent heat of evaporation, 
        :math:`\lambda_{24}`
        [J/kg]

    Returns
    -------
    t_24_init_mm : float
        initial estimate daily transpiration in mm, 
        :math:`T^{0}`
        [mm d-1]
    """
    return t_24_init * c.day_sec / lh_24



# -*- coding: utf-8 -*-
"""
Authors: Laust Færch
Module: Collect/PROBAV

Description:
This module preprocesses Landsat(7+8) data downloaded externaly

"""
from pywapor.collect.product import Landsat, CHIRPS, ERA5, GEOS5, GLOBCOVER, MERRA2, MODIS, PROBAV, SRTM, STATICS, SENTINEL2, SENTINEL3, VIIRSL1, COPERNICUS

__all__ = ["Landsat", "CHIRPS", "ERA5", "GEOS5", "GLOBCOVER", "MERRA2", "MODIS", "PROBAV", "SRTM", "STATICS", "SENTINEL2", "SENTINEL3", "VIIRSL1", "COPERNICUS"]

__version__ = '0.1'
